from os import path, sep

with open("{p}{s}VERSION".format(p=path.abspath(path.dirname(__file__)),
                                 s=sep)) as f:
    version = f.read().strip('\n')

defaults = {
    'DEFAULT_REQUEST_TIMEOUT': 30,
    'FLUSH_INTERVAL_SECS': 0.25,
    'FLUSH_LIMIT': 2 * 1024 * 1024,
    'MAX_CONCURRENT_REQUESTS': 10,
    'MAX_RETRY_ATTEMPTS': 3,
    'MAX_RETRY_JITTER': 0.5,
    'META_FIELDS': ['args', 'name', 'pathname', 'lineno'],
    'LOGDNA_URL': 'https://logs.logdna.com/logs/ingest',
    'BUF_RETENTION_LIMIT': 4 * 1024 * 1024,
    'RETRY_INTERVAL_SECS': 5,
    'USER_AGENT': 'python/%s' % version
}



import logging
import requests
import socket
import sys
import threading
import time

from concurrent.futures import ThreadPoolExecutor

from .configs import defaults
from .utils import sanitize_meta, get_ip, normalize_list_option


class LogDNAHandler(logging.Handler):
    def __init__(self, key, options={}):
        # Setup Handler
        logging.Handler.__init__(self)

        # Set Internal Logger
        self.internal_handler = logging.StreamHandler(sys.stdout)
        self.internal_handler.setLevel(logging.DEBUG)
        self.internalLogger = logging.getLogger('internal')
        self.internalLogger.addHandler(self.internal_handler)
        self.internalLogger.setLevel(logging.DEBUG)

        # Set the Custom Variables
        self.key = key
        self.hostname = options.get('hostname', socket.gethostname())
        self.ip = options.get('ip', get_ip())
        self.mac = options.get('mac', None)
        self.loglevel = options.get('level', 'info')
        self.app = options.get('app', '')
        self.env = options.get('env', '')
        self.tags = normalize_list_option(options, 'tags')
        self.custom_fields = normalize_list_option(options, 'custom_fields')
        self.custom_fields += defaults['META_FIELDS']

        # Set the Connection Variables
        self.url = options.get('url', defaults['LOGDNA_URL'])
        self.request_timeout = options.get('request_timeout',
                                           defaults['DEFAULT_REQUEST_TIMEOUT'])
        self.user_agent = options.get('user_agent', defaults['USER_AGENT'])
        self.max_retry_attempts = options.get('max_retry_attempts',
                                              defaults['MAX_RETRY_ATTEMPTS'])
        self.max_retry_jitter = options.get('max_retry_jitter',
                                            defaults['MAX_RETRY_JITTER'])
        self.max_concurrent_requests = options.get(
            'max_concurrent_requests', defaults['MAX_CONCURRENT_REQUESTS'])
        self.retry_interval_secs = options.get('retry_interval_secs',
                                               defaults['RETRY_INTERVAL_SECS'])

        # Set the Flush-related Variables
        self.buf = []
        self.buf_size = 0
        self.secondary = []
        self.exception_flag = False
        self.flusher = None
        self.include_standard_meta = options.get('include_standard_meta', None)

        if self.include_standard_meta is not None:
            self.internalLogger.debug(
                '"include_standard_meta" option will be deprecated ' +
                'removed in the upcoming major release')

        self.index_meta = options.get('index_meta', False)
        self.flush_limit = options.get('flush_limit', defaults['FLUSH_LIMIT'])
        self.flush_interval_secs = options.get('flush_interval',
                                               defaults['FLUSH_INTERVAL_SECS'])
        self.buf_retention_limit = options.get('buf_retention_limit',
                                               defaults['BUF_RETENTION_LIMIT'])

        # Set up the Thread Pools
        self.worker_thread_pool = ThreadPoolExecutor()
        self.request_thread_pool = ThreadPoolExecutor(
            max_workers=self.max_concurrent_requests)

        self.setLevel(logging.DEBUG)
        self.lock = threading.RLock()

    def start_flusher(self):
        if not self.flusher:
            self.flusher = threading.Timer(self.flush_interval_secs,
                                           self.flush)
            self.flusher.start()

    def close_flusher(self):
        if self.flusher:
            self.flusher.cancel()
            self.flusher = None

    def buffer_log(self, message):
        if self.worker_thread_pool:
            try:
                self.worker_thread_pool.submit(self.buffer_log_sync, message)
            except RuntimeError:
                self.buffer_log_sync(message)
            except Exception as e:
                self.internalLogger.debug('Error in calling buffer_log: %s', e)

    def buffer_log_sync(self, message):
        # Attempt to acquire lock to write to buf
        # otherwise write to secondary as flush occurs
        if self.lock.acquire(blocking=False):
            msglen = len(message['line'])
            if self.buf_size + msglen < self.buf_retention_limit:
                self.buf.append(message)
                self.buf_size += msglen
            else:
                self.internalLogger.debug(
                    'The buffer size exceeded the limit: %s',
                    self.buf_retention_limit)
            self.lock.release()

            if self.buf_size >= self.flush_limit and not self.exception_flag:
                self.close_flusher()
                self.flush()
            else:
                self.start_flusher()
        else:
            self.secondary.append(message)

    def clean_after_success(self):
        del self.buf[:]
        self.buf_size = 0
        self.exception_flag = False
        self.close_flusher()

    def flush(self):
        if self.worker_thread_pool:
            try:
                self.worker_thread_pool.submit(self.flush_sync)
            except RuntimeError:
                self.flush_sync()
            except Exception as e:
                self.internalLogger.debug('Error in calling flush: %s', e)

    def flush_sync(self):
        if self.buf_size == 0:
            return
        if self.lock.acquire(blocking=False):
            if self.request_thread_pool:
                try:
                    self.request_thread_pool.submit(self.try_request)
                except RuntimeError:
                    self.try_request()
                except Exception as e:
                    self.internalLogger.debug(
                        'Error in calling try_request: %s', e)
            self.lock.release()
        else:
            self.start_flusher()

    def try_request(self):
        self.buf.extend(self.secondary)
        self.secondary = []
        data = {'e': 'ls', 'ls': self.buf}
        retries = 0
        while retries < self.max_retry_attempts:
            retries += 1
            if self.send_request(data):
                self.clean_after_success()
                break

            sleep_time = self.retry_interval_secs * (1 << (retries - 1))
            sleep_time += self.max_retry_jitter
            time.sleep(sleep_time)

        if retries >= self.max_retry_attempts:
            self.internalLogger.debug(
                'Flush exceeded %s tries. Discarding flush buffer',
                self.max_retry_attempts)
            self.close_flusher()
            self.exception_flag = True

    def send_request(self, data):
        try:
            response = requests.post(url=self.url,
                                     json=data,
                                     auth=('user', self.key),
                                     params={
                                         'hostname': self.hostname,
                                         'ip': self.ip,
                                         'mac': self.mac,
                                         'tags': self.tags
                                     },
                                     stream=True,
                                     timeout=self.request_timeout,
                                     headers={'user-agent': self.user_agent})

            response.raise_for_status()
            status_code = response.status_code
            if status_code in [401, 403]:
                self.internalLogger.debug(
                    'Please provide a valid ingestion key.' +
                    ' Discarding flush buffer')
                return True

            if status_code == 200:
                return True

            if status_code in [400, 500, 504]:
                self.internalLogger.debug('The request failed %s. Retrying...',
                                          response.reason)
                return True
            else:
                self.internalLogger.debug(
                    'The request failed: %s. Retrying...', response.reason)

        except requests.exceptions.Timeout as timeout:
            self.internalLogger.debug('Timeout error occurred %s. Retrying...',
                                      timeout)

        except requests.exceptions.RequestException as exception:
            self.internalLogger.debug(
                'Error sending logs %s. Discarding flush buffer', exception)
            return True

        return False

    def emit(self, record):
        msg = self.format(record)
        record = record.__dict__
        message = {
            'hostname': self.hostname,
            'timestamp': int(time.time() * 1000),
            'line': msg,
            'level': record['levelname'] or self.loglevel,
            'app': self.app or record['module'],
            'env': self.env
        }

        message['meta'] = {}
        for key in self.custom_fields:
            if key in record:
                if isinstance(record[key], tuple):
                    message['meta'][key] = list(record[key])
                elif record[key] is not None:
                    message['meta'][key] = record[key]

        message['meta'] = sanitize_meta(message['meta'], self.index_meta)
        self.buffer_log(message)

    def close(self):
        self.close_flusher()
        self.flush_sync()
        if self.worker_thread_pool:
            self.worker_thread_pool.shutdown(wait=True)
            self.worker_thread_pool = None
        if self.request_thread_pool:
            self.request_thread_pool.shutdown(wait=True)
            self.request_thread_pool = None
        logging.Handler.close(self)



import unittest
from unittest.mock import patch
from logdna.utils import is_jsonable
from logdna.utils import sanitize_meta
from logdna.utils import get_ip
from logdna.utils import normalize_list_option

IP = '10.0.50.10'
VIP = '10.1.60.20'


class JSONTest(unittest.TestCase):
    def setUp(self):
        self.valid = {'key': 'value'}
        self.invalid = {'key': set()}

    def test_serialize_valid_json(self):
        self.assertTrue(is_jsonable(self.valid), 'json serializeble = True')

    def test_serialize_invalid_json(self):
        self.assertFalse(is_jsonable(self.invalid),
                         'non json serializable = False')


class SanitizeTest(unittest.TestCase):
    def setUp(self):
        self.valid = {'foo': 'bar', 'baz': 'whizbang'}
        self.invalid = {'bar': 'foo', 'baz': set()}

    def test_sanitize_simple(self):
        clean = sanitize_meta(self.valid, True)
        self.assertDictEqual(clean, self.valid)

    def test_sanitize_complex(self):
        clean = sanitize_meta(self.invalid, True)
        self.assertDictEqual(clean, {
            'bar': 'foo',
            '__errors': 'These keys have been sanitized: baz'
        })


class IPTest(unittest.TestCase):
    @patch('socket.socket', **{'return_value.connect.side_effect': OSError()})
    def test_get_ip_socket_error(self, _):
        self.assertEqual(get_ip(), '127.0.0.1',
                         'default to localhost on error')

    @patch('socket.socket',
           **{'return_value.getsockname.return_value': [IP, VIP]})
    def test_get_ip_default(self, _):
        self.assertEqual(get_ip(), IP, 'default to localhost on error')


class NormalizeListOptionTest(unittest.TestCase):
    def test_normalize_simple(self):
        value1 = normalize_list_option({'tags': ' a, b'}, 'tags')
        value2 = normalize_list_option({'tags': ['a', 'b']}, 'tags')
        value3 = normalize_list_option({'tags': ('a', 'b')}, 'tags')
        self.assertEqual(value1, ['a', 'b'])
        self.assertEqual(value1, value2)
        self.assertEqual(value3, [])



import logging
import unittest
import requests

from logdna import LogDNAHandler
from concurrent.futures import ThreadPoolExecutor
from logdna.configs import defaults

expectedLines = []
LOGDNA_API_KEY = '< YOUR INGESTION KEY HERE >'
logger = logging.getLogger('logdna')
logger.setLevel(logging.INFO)
sample_record = logging.LogRecord('test', logging.INFO, 'test', 5,
                                  'Something to test', '', '', '', '')
sample_message = {
    'line': 'Something to test',
    'hostname': 'localhost',
    'level': 'INFO',
    'app': 'test',
    'env': '',
    'meta': {
        'args': '',
        'name': 'test',
        'pathname': 'test',
        'lineno': 5
    }
}
sample_options = {
    'hostname': 'localhost',
    'ip': '10.0.1.1',
    'mac': 'C0:FF:EE:C0:FF:EE',
    'tags': 'sample,test',
    'index_meta': True
}


class MockThreadPoolExecutor():
    def __init__(self, **kwargs):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        pass

    def submit(self, fn, *args, **kwargs):
        # execute functions in series without creating threads
        # for easier unit testing
        result = fn(*args, **kwargs)
        return result

    def shutdown(self, wait=True):
        pass


class LogDNAHandlerTest(unittest.TestCase):
    def handler_test(self):
        handler = LogDNAHandler(LOGDNA_API_KEY, sample_options)
        self.assertIsInstance(handler, logging.Handler)
        self.assertIsInstance(handler.internal_handler, logging.StreamHandler)
        self.assertIsNotNone(handler.internalLogger)
        self.assertEqual(handler.key, LOGDNA_API_KEY)
        self.assertEqual(handler.hostname, sample_options['hostname'])
        self.assertEqual(handler.ip, sample_options['ip'])
        self.assertEqual(handler.mac, sample_options['mac'])
        self.assertEqual(handler.loglevel, 'info')
        self.assertEqual(handler.app, '')
        self.assertEqual(handler.env, '')
        self.assertEqual(handler.tags, sample_options['tags'].split(','))
        self.assertEqual(handler.custom_fields, defaults['META_FIELDS'])

        # Set the Connection Variables
        self.assertEqual(handler.url, defaults['LOGDNA_URL'])
        self.assertEqual(handler.request_timeout,
                         defaults['DEFAULT_REQUEST_TIMEOUT'])
        self.assertEqual(handler.user_agent, defaults['USER_AGENT'])
        self.assertEqual(handler.max_retry_attempts,
                         defaults['MAX_RETRY_ATTEMPTS'])
        self.assertEqual(handler.max_retry_jitter,
                         defaults['MAX_RETRY_JITTER'])
        self.assertEqual(handler.max_concurrent_requests,
                         defaults['MAX_CONCURRENT_REQUESTS'])
        self.assertEqual(handler.retry_interval_secs,
                         defaults['RETRY_INTERVAL_SECS'])

        # Set the Flush-related Variables
        self.assertEqual(handler.buf, [])
        self.assertEqual(handler.buf_size, 0)
        self.assertEqual(handler.secondary, [])
        self.assertFalse(handler.exception_flag)
        self.assertIsNone(handler.flusher)
        self.assertTrue(handler.index_meta)
        self.assertEqual(handler.flush_limit, defaults['FLUSH_LIMIT'])
        self.assertEqual(handler.flush_interval_secs,
                         defaults['FLUSH_INTERVAL_SECS'])
        self.assertEqual(handler.buf_retention_limit,
                         defaults['BUF_RETENTION_LIMIT'])

        # Set up the Thread Pools
        self.assertIsInstance(handler.worker_thread_pool, ThreadPoolExecutor)
        self.assertIsInstance(handler.request_thread_pool, ThreadPoolExecutor)
        self.assertEqual(handler.level, logging.DEBUG)

    def flusher_test(self):
        handler = LogDNAHandler(LOGDNA_API_KEY, sample_options)
        self.assertIsNone(handler.flusher)
        handler.start_flusher()
        self.assertIsNotNone(handler.flusher)
        handler.close_flusher()
        self.assertIsNone(handler.flusher)

    def emit_test(self):
        handler = LogDNAHandler(LOGDNA_API_KEY, sample_options)
        handler.buffer_log = unittest.mock.Mock()
        handler.emit(sample_record)
        sample_message['timestamp'] = unittest.mock.ANY
        handler.buffer_log.assert_called_once_with(sample_message)

    def try_request_test(self):
        requests.post = unittest.mock.Mock()
        handler = LogDNAHandler(LOGDNA_API_KEY, sample_options)
        sample_message['timestamp'] = unittest.mock.ANY
        handler.buf = [sample_message]
        handler.try_request()
        requests.post.assert_called_with(
            url=handler.url,
            json={
                'e': 'ls',
                'ls': handler.buf
            },
            auth=('user', handler.key),
            params={
                'hostname': handler.hostname,
                'ip': handler.ip,
                'mac': handler.mac,
                'tags': handler.tags
            },
            stream=True,
            timeout=handler.request_timeout,
            headers={'user-agent': handler.user_agent})

    def close_test(self):
        handler = LogDNAHandler(LOGDNA_API_KEY, sample_options)
        handler.close_flusher = unittest.mock.Mock()
        handler.flush_sync = unittest.mock.Mock()
        handler.close()
        handler.close_flusher.assert_called_once_with()
        handler.flush_sync.assert_called_once_with()
        self.assertIsNone(handler.worker_thread_pool)
        self.assertIsNone(handler.request_thread_pool)

    def flush_test(self):
        handler = LogDNAHandler(LOGDNA_API_KEY, sample_options)
        handler.worker_thread_pool = MockThreadPoolExecutor()
        handler.request_thread_pool = MockThreadPoolExecutor()
        handler.buf_size += 1
        handler.try_request = unittest.mock.Mock()
        handler.flush()
        handler.try_request.assert_called_once_with()

    def buffer_log_test(self):
        handler = LogDNAHandler(LOGDNA_API_KEY, sample_options)
        handler.worker_thread_pool = MockThreadPoolExecutor()
        handler.request_thread_pool = MockThreadPoolExecutor()
        handler.close_flusher = unittest.mock.Mock()
        handler.flush = unittest.mock.Mock()
        sample_message['timestamp'] = unittest.mock.ANY
        handler.flush_limit = 0
        handler.buffer_log(sample_message)
        handler.close_flusher.assert_called_once_with()
        handler.flush.assert_called_once_with()
        self.assertEqual(handler.buf, [sample_message])
        self.assertEqual(handler.buf_size, len(sample_message['line']))

    def test_run_tests(self):
        self.handler_test()
        self.flusher_test()
        self.emit_test()
        self.try_request_test()
        self.close_test()
        self.flush_test()
        self.buffer_log_test()


if __name__ == '__main__':
    unittest.main()
